import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
# set big font
import seaborn as sns
sns.set_context("notebook", font_scale=1.8)
plt.style.use('fivethirtyeight')
import timeit
import datetime
import itertools
import argparse
from sklearn.preprocessing import MinMaxScaler
from sklearn.metrics import roc_auc_score
from sklearn.model_selection import train_test_split
from keras.layers import Input, Dense, Concatenate, Lambda, Dot
from keras.models import Model, Sequential
from keras import regularizers
from keras import backend as K
from keras import optimizers

import os
os.environ['CUDA_VISIBLE_DEVICES'] = '-1'
import tensorflow as tf
config = tf.ConfigProto()
config.gpu_options.allow_growth = True
sess = tf.Session(config=config)
K.set_session(sess)

parser = argparse.ArgumentParser()
parser.add_argument('--datasource', default="1", type=str, nargs='?', help='datasource')
args = parser.parse_args()
print("datasource={}".format(args.datasource))

np.random.seed(123)

datasource = args.datasource
if datasource == "1":
    dataset_all = ["1a-selbal-crohn", "1b-selbal-msm-hivonly"]
elif datasource == "2":
    dataset_all = ["2a-franzosa-mic-ibd-v-hc", "2b-franzosa-mic-cd-v-uc",
                   "2c-franzosa-met10-ibd-v-hc", "2d-franzosa-met10-cd-v-uc"]
elif datasource == "3":
    dataset_all = ["3a-duvallet-schubert-case-v-diarrhea", "3b-duvallet-schubert-case-v-nondc",
                   "3c-duvallet-baxter-crc-v-h", "3d-duvallet-baxter-crc-v-noncrc"]
elif datasource == "4":
    dataset_all = ["4a-brca-mirna10-ca-v-hc", "4b-brca-mirna10-her2-v-ca", "4c-brca-mirna10-lumA-v-lumB"]

test_size = 0.1
sumzero_lambda = 1e0
l1_lambdas = [1e-3, 1e-2, 1e-1, 1e0]
use_weight_constraint = True
n_run = 10

# hyper-parameters for network
cascade_levels = [1, 3, 5, 10]
bottle_dim = 1
hidden_dim = 32
latent_dim = 1
output_dim = 1
batch_size = 32
epochs = 200 # 200

# regularize sum of weights at each cascade to be 0 and regularize weights to be sparse
class SumZeroL1Reg(regularizers.Regularizer):
    def __init__(self, sumzero_lambda=1e0, l1_lambda=1e-3):
        self.sumzero_lambda = K.cast_to_floatx(sumzero_lambda)
        self.l1_lambda = K.cast_to_floatx(l1_lambda)

    def __call__(self, w):
        sumzero_reg = 0
        sumzero_reg += self.sumzero_lambda * K.square(K.sum(w))

        l1_reg = 0
        l1_reg += self.l1_lambda * K.sum(K.abs(w))

        return sumzero_reg + l1_reg

    def get_config(self):
        return {'sumzero_lambda': float(self.sumzero_lambda),
                'l1_lambda': float(self.l1_lambda)}

cascade_l1_rng = [cascade_levels, l1_lambdas]
all_cascade_l1_values = []
for element in itertools.product(*cascade_l1_rng):
    all_cascade_l1_values.append(element)
all_cascade_l1_values = np.array(all_cascade_l1_values)

auc_dataset_para_run = []
for dataset in dataset_all:
    # read data from file
    df_x = pd.read_csv("./data/{}-x.csv".format(dataset), header=0, sep=",")
    df_y = pd.read_csv("./data/{}-y.csv".format(dataset), header=0, sep=",")
    X = df_x.iloc[:, 1:]
    y = df_y.iloc[:, 1]
    X = np.array(X)
    y = np.array(y)
    y = y.reshape(-1, 1)
    n_sample = X.shape[0]
    n_feature = X.shape[1]
    print("dataset={}, n_samples={}, n_features={}".format(dataset, n_sample, n_feature))
    # scale X to [a, b]
    X = MinMaxScaler((1, 2)).fit_transform(X)

    auc_para_run = []
    for cascade_l1 in all_cascade_l1_values:
        start_date_time = datetime.datetime.now()
        start_time = timeit.default_timer()

        cascade_level = int(cascade_l1[0])
        l1_lambda = cascade_l1[1]
        print("cascade_lvl={}, sumzero_lambda={}, l1_lambda={}".format(cascade_level, sumzero_lambda, l1_lambda))

        auc_run = np.zeros(n_run)
        for run in range(n_run):
            print("run={}".format(run))
            # split data to train and test sets
            X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=test_size, random_state=run, stratify=y)
            # construct network
            time1 = timeit.default_timer()
            input_dim = n_feature
            # Encoder
            x = Input(shape=(input_dim,))
            # concat layer for all z
            concat_z = []
            for level_id in range(cascade_level):
                x_log = Lambda(lambda t: K.log(t))(x)
                if use_weight_constraint == True:
                    b = Dense(bottle_dim, activation='linear',
                              kernel_regularizer=SumZeroL1Reg(l1_lambda=l1_lambda))(x_log)
                else:
                    b = Dense(bottle_dim, activation='linear')(x_log)
                z = b
                concat_z.append(z)
            if cascade_level == 1:
                all_z = z
            else:
                all_z = Concatenate()(concat_z)
            h = Dense(hidden_dim, activation='relu')(all_z)
            beta = Dense(cascade_level, activation='relu')(h)
            # Decoder
            all_z_beta = Dot(axes=1)([all_z, beta])
            decoder = Sequential([Dense(output_dim, input_dim=output_dim, activation='sigmoid')])
            y_pred = decoder(all_z_beta)
            # y_pred = Dot(axes=1)([all_z, beta])
            # train network
            model = Model(inputs=x, outputs=y_pred, name='bottleneck_model')
            opt = optimizers.Adam()
            model.compile(optimizer=opt, loss='binary_crossentropy', metrics=['accuracy'])
            hist = model.fit(X_train, y_train, shuffle=True, epochs=epochs, batch_size=batch_size, validation_split=0.1, verbose=0)
            time2 = timeit.default_timer()
            print("runtime of training network: {}(s)".format(round(time2 - time1, 2)))

            # compute auc on test set
            y_test_pred = model.predict(X_test)
            y_test_pred_round = np.around(y_test_pred)
            auc = roc_auc_score(y_test, y_test_pred)
            print("auc={}".format(round(auc, 4)))
            auc_run[run] = auc
            """
            # plot loss of training network
            fig_size = lambda width: (width, 2. * width / (1 + np.sqrt(5)))
            fig, ax = plt.subplots(figsize=fig_size(6))
            hist_df = pd.DataFrame(hist.history)
            hist_df = hist_df[['loss', 'val_loss']]
            hist_df.plot(ax=ax)
            ax.set_ylabel('loss')
            ax.set_xlabel('# epochs')
            ax.set_ylim(.99 * hist_df[1:].values.min(), 1.1 * hist_df[1:].values.max())
            plt.savefig('./{}_loss_lvl{}_regsum{}_regl1{}_run{}.pdf'.format(dataset, cascade_level, sumzero_lambda, l1_lambda, run),
                        bbox_inches="tight")
            plt.close()
            """
            # delete trained network to free memory
            del model, decoder
            K.clear_session()

        end_date_time = datetime.datetime.now()
        end_time = timeit.default_timer()
        runtime = round(end_time-start_time, 2)

        auc_run = np.around(auc_run, 4)
        avg_auc = round(np.mean(auc_run), 4)
        std_err = round(np.std(auc_run) / np.sqrt(n_run), 2)
        print("dataset={}, cascade_lvl={}, sumzero_lambda={}, l1_lambda={}".format(dataset, cascade_level, sumzero_lambda, l1_lambda))
        print("auc_run=[{}]".format(",".join(map(str, auc_run))))
        print("avg_auc={}, std_err={}".format(avg_auc, std_err))
        print("start date time: {} and end date time: {}".format(start_date_time, end_date_time))
        print("runtime: {}(s)".format(runtime))
        # save result to file
        with open('with_attention_{}_lvl_{}_l1_{}.txt'.format(dataset, cascade_level, l1_lambda), 'w') as f:
            f.write("dataset={}, cascade_lvl={}, sumzero_lambda={}, l1_lambda={}\n".format(dataset, cascade_level, sumzero_lambda, l1_lambda))
            f.write("auc_run=[{}]\n".format(",".join(map(str, auc_run))))
            f.write("avg_auc={}, std_err={}\n".format(avg_auc, std_err))
            f.write("runtime: {}(s)".format(runtime))
        # store auc of n_run of each parameter
        auc_para_run.append(auc_run)
    # store auc of n_run of all parameters of each dataset
    auc_dataset_para_run.append(auc_para_run)

# save result to csv file
n_dataset = len(dataset_all)
n_para = len(all_cascade_l1_values)
with open('with_attention_data_source_{}.csv'.format(datasource), 'w') as f:
    f.write("dataset, cascade_lvl, sumzero_lambda, l1_lambda, bootstrapp, auc\n")
    for data_id in range(n_dataset):
        for para_id in range(n_para):
            for auc_id in range(n_run):
                data_name = dataset_all[data_id]
                cascade_l1 = all_cascade_l1_values[para_id]
                cascade_level = int(cascade_l1[0])
                l1_lambda = cascade_l1[1]
                line = data_name + "," + str(cascade_level) + "," + str(sumzero_lambda) + "," + str(l1_lambda) + "," + \
                       str(auc_id) + "," + str(auc_dataset_para_run[data_id][para_id][auc_id]) + "\n"
                f.write(line)

